#!/usr/bin/ruby -Ku

# EUC-JISX0213 又は Shift_JISX0213で符号化されたplain textを
# sxw形式に変換する

# ruby txt2sxw.rb [-e] [-s] [-r] [-f <font_name>] <infile> <outfile>
#  ただし -e は EUC、-s はSJIS、-r は青空文庫形式のルビをsxwのルビに変換。
#  -e と -s を同時には指定できない。デフォルトは -e。
#  オプションは -sr のように表記可能。

require 'getopts'
require 'iconv'
require 'zip/zipfilesystem'

def utf8ToScalar(str)
  if str.length == 1
    return str[0]
  elsif str.length == 2
    firstByte = str[0] & 0b00011111
    secondByte = str[1] & 0b00111111
    return (firstByte << 6) + secondByte
  elsif str.length == 3
    firstByte = str[0] & 0b00001111
    secondByte = str[1] & 0b00111111
    thirdByte = str[2] & 0b00111111
    return (firstByte << 12) + (secondByte << 6) + thirdByte
  elsif str.length == 4
    firstByte = str[0] & 0b00000111
    secondByte = str[1] & 0b00111111
    thirdByte = str[2] & 0b00111111
    fourthByte = str[3] & 0b00111111
    return (firstByte << 18) + (secondByte << 12) + (thirdByte << 6) + fourthByte
  elsif str.length == 0
    return 0
  end
  return -1
end

def isKanji?(scalar)
  return (scalar == 0x3005) || (scalar >= 0x3400 && scalar <= 0x4dbf) || (scalar >= 0x4e00 && scalar <= 0x9fff) || (scalar >= 0xf900 && scalar <= 0xfaff) || (scalar >= 0x20000 && scalar <= 0x2a6df) || (scalar >= 0x2f800 && scalar <= 0x2fa1f)
end
  
def isLatin?(scalar)
  return (scalar == 0x20) || (scalar >= 0x30 && scalar <= 0x39) || (scalar >= 0x41 && scalar <= 0x5a) || (scalar >= 0x61 && scalar <= 0x7a)
end

RubyStartChar = "《"  # U+300A
RubyEndChar = "》"    # U+300B
RubyBaseStartChar = "｜"  # U+FF5C, 'fullwidth' version of '|'

def markUpLine(line)
  result = ""
  rubyBaseCandidate = ""
  inRuby = false
  rubyString = ""
  explicitlyDelimited = false
  line.scan(/./) do |char|
    if !inRuby
      scalar = utf8ToScalar(char)
      if isKanji?(scalar) || isLatin?(scalar)
        rubyBaseCandidate.concat(char)
      elsif char == RubyBaseStartChar
        result.concat(rubyBaseCandidate)
        rubyBaseCandidate = ""
        explicitlyDelimited = true
      elsif !(char == RubyStartChar)
        if explicitlyDelimited
          result.concat(RubyBaseStartChar)
        end
        result.concat(rubyBaseCandidate)
        result.concat(char)
        rubyBaseCandidate = ""
        explicitlyDelimited = false
      else
        if rubyBaseCandidate.length > 0
          rubyString = ""
          inRuby = true
        else
          if explicitlyDelimited
            result.concat(RubyBaseStartChar)
          end
          result.concat(RubyStartChar)
          explicitlyDelimited = false
        end
      end
    else
      if char == RubyEndChar
        if rubyString.length > 0
          result.concat('<text:ruby text:style-name="Ru1"><text:ruby-base>')
          result.concat(rubyBaseCandidate)
          result.concat('</text:ruby-base><text:ruby-text>')
          result.concat(rubyString)
          result.concat('</text:ruby-text></text:ruby>')
        else
          result.concat(RubyStartChar)
          result.concat(RubyEndChar)
        end
        rubyBaseCandidate = ""
        explicitlyDelimited = false
        inRuby = false
      else
        rubyString.concat(char)
      end
    end
  end
  if explicitlyDelimited
    result.concat(RubyBaseStartChar)
  end
  if inRuby
    result.concat(rubyBaseCandidate)
    result.concat(RubyStartChar)
    result.concat(rubyString)
  else
    if rubyBaseCandidate.length > 0
      result.concat(rubyBaseCandidate)
    end
  end
  return result
end

# main

VERSION_STRING = "0.2"
USAGE_STRING = "usage: #$0 [-esrv] [--version] [-f font] infile outfile"

unless getopts('esrv', 'f:', 'version')
  abort USAGE_STRING
end

if $OPT_v or $OPT_version
  puts VERSION_STRING
  exit
end

fontname = "HG 明朝L Sun"
if $OPT_f
  arg_code = (ENV["OS"] =~ /Windows/) ? "Shift_JISX0213" : "EUC-JISX0213"
  argconv = Iconv.new("UTF-8", arg_code)
  fontname = argconv.iconv($OPT_f)
end

use_ruby = $OPT_r

from_code = ""
if $OPT_s
  if $OPT_e
    abort "#$0: error: both -e and -s cannot be specified at once."
  end
  from_code = "Shift_JISX0213"
else
  from_code = "EUC-JISX0213"
end

infile = ARGV.shift
if infile.nil?
  abort USAGE_STRING
end

outfile = ARGV.shift
if outfile.nil?
  abort USAGE_STRING
end

# 入力テキストをUTF-8に変換する

origtxt = ""
open(infile) do |f|
  while line = f.gets
    origtxt.concat(line)
  end
end

begin
  converter = Iconv.new("UTF-8", from_code)
  utf8txt = converter.iconv(origtxt)
rescue Iconv::IllegalSequence
  puts "#$0: error: Input file cannot be interpreted as " + from_code
  exit
end

# 規定のXMLファイルにテキストを埋め込み、zipで圧縮する

open(outfile, "w") do |to|
  open(File::dirname($0) + "/template.sxw") do |from|
    fragment = from.read
    to.write fragment
  end
end

markedUpText = ""
utf8txt.each do |line|
  line.chomp!
  c = 0;
  line.each_byte do |byte|
    if (byte == 0x20)
      c += 1;
    else
      break
    end
  end
  line.slice!(0, c)
  markedUpText.concat('<text:p text:style-name="P1">')
  if (c > 0)
    markedUpText.concat('<text:s text:c="' + c.to_s + '"/>')
  end
  markedUpText.concat(use_ruby ? markUpLine(line) : line)
  markedUpText.concat("</text:p>")
end

xml_decl = ""
content = ""
Zip::ZipFile.open(outfile) do |zf|
  zf.file.open("content.xml") do |os|
    xml_decl = os.gets
    content = os.gets
    if fontname != nil
      content.gsub!(/!!!FONT-NAME!!!/, fontname)
    end
    content.sub!(/<!--text:p-->/, markedUpText)
  end
  zf.file.open("content.xml", "w") do |os|
    os.write(xml_decl)
    os.write(content)
  end
end
